"""This module contains generic input validation methods."""
from __future__ import annotations

import warnings
from numbers import Integral, Real
from typing import Any

import networkx as nx
import numpy as np
from numpy.typing import ArrayLike, NDArray

# pylint: disable=invalid-name


def check_real(
    x: Any,
    name: str,
    *,
    l_bound: float | None = None,
    u_bound: float | None = None,
    l_inclusive: bool = True,
    u_inclusive: bool = True,
) -> float:
    """Check if the variable `x` with name 'name' is a real number.

    Optionally, lower and upper bounds can also be checked.

    Args:
        x: Variable to check.
        name: Name of the variable. This name will be displayed in possible error
            messages.
        l_bound: Lower bound of `x`.
        u_bound: Upper bound of `x`.
        l_inclusive: If ``True`` the lower bound is inclusive, otherwise the lower bound
            is exclusive.
        u_inclusive: If ``True`` the upper bound is inclusive, otherwise the upper bound
            is exclusive.

    Raises:
        TypeError: If `x` is not a real number.
        ValueError: If `x` is outside the give bounds.

    Returns:
        Floating point representation of `x`.
    """
    if not isinstance(x, Real):
        raise TypeError(f"'{name}' should be a real number, but was of type {type(x)}")
    x_float = float(x)
    error_msg = "'" + name + "' has an {} {} bound of {}, but was " + str(x)
    if l_bound is not None:
        if l_inclusive:
            if x_float < l_bound:
                raise ValueError(error_msg.format("inclusive", "lower", l_bound))
        else:
            if x_float <= l_bound:
                raise ValueError(error_msg.format("exclusive", "lower", l_bound))

    if u_bound is not None:
        if u_inclusive:
            if x_float > u_bound:
                raise ValueError(error_msg.format("inclusive", "upper", u_bound))
        else:
            if x_float >= u_bound:
                raise ValueError(error_msg.format("exclusive", "upper", u_bound))

    return x_float


def check_int(
    x: Any,
    name: str,
    *,
    l_bound: float | None = None,
    u_bound: float | None = None,
    l_inclusive: bool = True,
    u_inclusive: bool = True,
) -> int:
    """Check if the variable `x` with name 'name' is a real number.

    Optionally, lower and upper bounds can also be checked.

    Args:
        x: Variable to check.
        name: Name of the variable. This name will be displayed in possible error
            messages.
        l_bound: Lower bound of `x`.
        u_bound: Upper bound of `x`.
        l_inclusive: If ``True`` the lower bound is inclusive, otherwise the lower
            bound is exclusive.
        u_inclusive: If ``True`` the upper bound is inclusive, otherwise the upper
            bound is exclusive.

    Raises:
        TypeError: If `x` is not a real number.
        ValueError: If `x` is outside the give bounds.

    Returns:
        Floating point representation of `x`.
    """
    if not isinstance(x, Real):
        raise TypeError(f"'{name}' should be an integer, but was of type {type(x)}")

    if not isinstance(x, Integral):
        int_x = int(x)
        if x - int_x != 0:
            msg = f"'{name}' with value {x} could not be safely converted to an integer"
            raise ValueError(msg)

    error_msg = "'" + name + "' has an {} {} bound of {}, but was " + str(x)
    x_int = int(x)
    if l_bound is not None:
        if l_inclusive:
            if x_int < l_bound:
                raise ValueError(error_msg.format("inclusive", "lower", l_bound))
        else:
            if x_int <= l_bound:
                raise ValueError(error_msg.format("exclusive", "lower", l_bound))

    if u_bound is not None:
        if u_inclusive:
            if x_int > u_bound:
                raise ValueError(error_msg.format("inclusive", "upper", u_bound))
        else:
            if x_int >= u_bound:
                raise ValueError(error_msg.format("exclusive", "upper", u_bound))

    return x_int


def check_string(x: str, name: str, *, lower: bool = False, upper: bool = False) -> str:
    """Check if the variable `x` with name 'name' is a string.

    Optionally, the string can be converted to all lowercase or all uppercase letters.

    Args:
        x: Variable to check.
        name: Name of the variable. This name will be displayed in possible error
            messages.
        lower: If ``True``, `x` will be returned with lowercase letters. Defaults to
            ``False``.
        upper: If ``True``, `x` will be returned with uppercase letters. Default to
            ``False``.

    Raises:
        TypeError: If `x` is not an instance of ``str``.

    Returns:
        Input string. Optionally, in lowercase or uppercase letters.
    """
    if not isinstance(x, str):
        raise TypeError(f"'{name}' must be a string, but was of type {type(x)}")

    if lower:
        x = x.lower()
    if upper:
        x = x.upper()
    return x


def check_bool(x: Any, name: str, *, safe: bool = False) -> bool:
    """Check if the variable `x` with name 'name' is a Boolean value.

    Optionally, cast to Boolean value if it is not.

    Args:
        x: Variable to check.
        name: Name of the variable. This name will be displayed in possible error
            messages.
        safe: If ``True`` raise a ``TypeError`` when `x` is not a bool. If ``False``
            cast to bool.

    Raises:
        TypeError: If `safe` is ``True`` and `x` is not a Boolean value.

    Returns:
        Boolean representation of the input.
    """
    if not isinstance(x, bool) and safe:
        raise TypeError(f"'{name}' must be a Boolean value, but was of type {type(x)}")
    return bool(x)


def check_adjacency_matrix(adjacency_matrix: ArrayLike) -> NDArray[Any]:
    """Check if a matrix is an adjacency matrix, i.e., a square matrix.

    Args:
        adjacency_matrix: Matrix to check.

    Raises:
        ValueError: When the provided input is not a square matrix.

    Returns:
        Square NDArray representation of ``adjacency_matrix``.
    """
    numpy_matrix: NDArray[Any]
    if hasattr(adjacency_matrix, "toarray"):
        numpy_matrix = adjacency_matrix.toarray()
    else:
        numpy_matrix = np.array(adjacency_matrix)

    if numpy_matrix.ndim != 2 or numpy_matrix.shape[0] != numpy_matrix.shape[1]:
        raise ValueError("The provided value should be a square 2-D adjacency matrix.")

    return numpy_matrix


def check_graph_is_valid_topology(graph: nx.Graph, name: str) -> None:
    """Check if `graph` with name 'name' is an instance of ``networkx.Graph`` and check
    if the graph is valid topology graph.

    Args:
        graph: Graph to check.
        name: Name of the graph. This name will be displayed in possible error
            messages.

    Raises:
        TypeError: If `graph` is not an instance of ``networkx.Graph``.
        ValueError: If `graph` is not a valid topology graph.
    """
    check_instance(graph, name, nx.Graph)

    if nx.number_of_selfloops(graph) > 0:
        raise ValueError(f"'{name}' contains self-loops")

    if len(graph) == 0:
        raise ValueError(f"'{name}' has no nodes")


def check_instance(x: Any, name: str, dtype: type) -> None:
    """Check if `x` with name 'name' is an instance of dtype.

    Args:
        x: Variable to check.
        name: Name of the variable. This name will be displayed in possible error
            messages.

    Raises:
        TypeError: If `x` is not an instance of `dtype`.
    """
    if not isinstance(x, dtype):
        msg = f"'{name}' must an instance of {dtype}, but was of type {type(x)}"
        raise TypeError(msg)


def warn_if_positive(x: float, name: str) -> None:
    """Give a warning when `x` is positive.

    Args:
        x: Variable to check.
        name: Name of the variable. This name will be displayed in the warning.
    """
    if x > 0:
        warnings.warn(f"'{name}' was positive")


def warn_if_negative(x: float, name: str) -> None:
    """Give a warning when `x` is negative.

    Args:
        x: Variable to check.
        name: Name of the variable. This name will be displayed in the warning.
    """
    if x < 0:
        warnings.warn(f"'{name}' was negative")



"""Generic abstract base class for States of RL environments.

All states should inherit from ``State``.
"""
from __future__ import annotations

from abc import abstractmethod
from typing import Any, Generic, TypeVar

from gymnasium.spaces import Space
from numpy.random import Generator, default_rng

ObservationT = TypeVar("ObservationT")
ActionT = TypeVar("ActionT")


class State(Generic[ObservationT, ActionT]):
    """RL State containing the current state of the problem."""

    steps_done: int
    """Number of steps done since the last reset."""
    _rng: Generator | None = None

    @abstractmethod
    def reset(
        self, *, seed: int | None = None, **_kwargs: Any
    ) -> State[ObservationT, ActionT]:
        """Reset the state.

        Returns:
            Self.
        """
        raise NotImplementedError

    def seed(self, seed: int | None = None) -> list[int | None]:
        """Seed the rng of this space, using ``numpy.random.default_rng``.

        Args:
            seed: Seed for the rng. Defaults to ``None``

        Returns:
            The used seeds.
        """
        self._rng = default_rng(seed)
        return [seed]

    @property
    def rng(self) -> Generator:
        """Return the random number generator of this environment. If none is set yet,
        this will generate a new one using ``numpy.random.default_rng``.

        Returns:
            Random number generator used by this ``Environment``.
        """
        if self._rng is None:
            self._rng = default_rng()
        return self._rng

    @rng.setter
    def rng(self, rng: Generator) -> None:
        self._rng = rng

    @abstractmethod
    def update_state(self, action: ActionT) -> State[ObservationT, ActionT]:
        """Update the state of this ``Environment`` using the given action.

        Args:
            action: Action to be executed.

        Returns:
            Self.
        """
        raise NotImplementedError

    @abstractmethod
    def obtain_observation(self) -> ObservationT:
        """Observation based on the current state."""
        raise NotImplementedError

    @abstractmethod
    def is_done(self) -> bool:
        """Boolean value stating whether we are in a final state."""
        raise NotImplementedError

    @abstractmethod
    def obtain_info(self) -> dict[Any, Any]:
        """Optional debugging info for the current state."""
        raise NotImplementedError

    @abstractmethod
    def create_observation_space(self) -> Space[Any]:
        """Create the corresponding observation space."""
        raise NotImplementedError

    def __repr__(self) -> str:
        text = f"{self.__class__.__name__}:\n"
        if hasattr(self, "__slots__"):
            for attribute_name in self.__slots__:
                text += f"{attribute_name}: {repr(getattr(self, attribute_name))}\n"
        if hasattr(self, "__dir__"):
            for attribute_name, attribute_value in self.__dict__.items():
                text += f"{attribute_name}: {repr(attribute_value)}\n"
        return text



"""This module contains the ``CommutationRulebook`` class together with basic
commutation rules used in the ``Scheduling`` environment.

Example:
    The code block below shows how to set up a ``CommutationRulebook``, with the
    additional commutation rule that two C-NOT gates with the same control qubit
    commute.

    .. code-block:: python

        from qgym.env.scheduling.rulebook import CommutationRulebook

        # cnot gates with the same control qubit commute
        def cnot_commutation(gate1, gate2):
            if gate1[0] == "cnot" and gate2[0] == "cnot":
                if gate1[1] == gate2[1]:
                    return True
            return False

        # init the rulebook and add the commutation rule
        rulebook = CommutationRulebook()
        rulebook.add_rule(rulebook)

"""
from __future__ import annotations

from typing import Callable

import numpy as np
from numpy.typing import NDArray

from qgym.custom_types import Gate


class CommutationRulebook:
    """Commutation rulebook used in the ``Scheduling`` environment."""

    def __init__(self, default_rules: bool = True) -> None:
        """Init of the ``CommutationRulebook``.

        Args:
            default_rules: If ``True``, default rules are used. Default rules dictate
                that gates with disjoint qubits commute and that gates that are exactly
                the same commute. If ``False``, then no rules will be initialized.
        """

        self._rules: list[Callable[[Gate, Gate], bool]]
        if default_rules:
            self._rules = [disjoint_qubits, same_gate]
        else:
            self._rules = []

    def make_blocking_matrix(self, circuit: list[Gate]) -> NDArray[np.int_]:
        """Make a square array of shape (len(circuit), len(circuit)), with dependencies
        based on the given commutation rules.

        Args:
            circuit: Circuit to check dependencies for.

        Returns:
            Dependencies matrix of the circuit based on the rules and scheduling from
            right to left.
        """
        blocking_matrix = np.zeros((len(circuit), len(circuit)), dtype=bool)

        for idx, gate in enumerate(circuit):
            for idx_other in range(idx + 1, len(circuit)):
                gate_other = circuit[idx_other]

                if not self.commutes(gate, gate_other):
                    blocking_matrix[idx, idx_other] = True

        return blocking_matrix

    def commutes(self, gate1: Gate, gate2: Gate) -> bool:
        """Check if `gate1` and `gate2` commute according to the rules in the rulebook.

        Args:
            gate1: Gate to check the commutation.
            gate2: Gate to check gate1 against.

        Returns:
            Boolean value indicating whether `gate1` commutes with `gate2`.
        """
        for rule in self._rules:
            if rule(gate1, gate2):
                return True
        return False

    def add_rule(self, rule: Callable[[Gate, Gate], bool]) -> None:
        """Add a new commutation rule to the rulebook.

        Args:
            rule: Rule to add to the rulebook. A rule is a ``Callable`` which takes as
                input two gates and returns a Boolean value that should be ``True`` if
                two gates commute and ``False`` otherwise.
        """
        self._rules.append(rule)

    def __repr__(self) -> str:
        """Create a string representation of the CommutationRulebook."""
        text = f"{self.__class__.__name__}(rules=["
        for rule in self._rules:
            if callable(rule) and hasattr(rule, "__name__"):
                text += f"{rule.__name__}, "
            else:
                text += f"{rule}, "
        text = text[:-2] + "])"
        return text


def disjoint_qubits(gate1: Gate, gate2: Gate) -> bool:
    """Gates that have disjoint qubits commute.

    Args:
        gate1: Gate to check disjointness.
        gate2: Gate to check disjointness against.

    Returns:
        Boolean value stating whether the gates are disjoint.
    """
    return bool(
        gate1.q1 != gate2.q1
        and gate1.q1 != gate2.q2
        and gate1.q2 != gate2.q1
        and gate1.q2 != gate2.q2
    )


def same_gate(gate1: Gate, gate2: Gate) -> bool:
    """Gates that are equal commute.

    Args:
        gate1: Gate to check equality.
        gate2: Gate to check equality against.

    Returns:
        Boolean value stating whether the gates are equal.
    """
    return gate1 == gate2



"""This module contains the ``MachineProperties`` class, which helps with conveniently
setting up machine properties for the scheduling environment.

Usage:
    The code below will create ``MachineProperties`` which will have the following
    properties:

    * The machine has two qubits.
    * The machine supports the X, Y, C-NOT and Measure gates.
    * Multiple Measure gates should start in the same cycle, or wait till the previous
      one is done.
    * The X and Y gates cannot be performed in the same cycle.

    .. code-block:: python

        from qgym.envs.scheduling import MachineProperties

        hardware_spec = MachineProperties(n_qubits=2)
        hardware_spec.add_gates({"x": 2, "y": 2, "cnot": 4, "measure": 10})
        hardware_spec.add_same_start(["measure"])
        hardware_spec.add_not_in_same_cycle([("x", "y")])


"""
from __future__ import annotations

import warnings
from copy import deepcopy
from typing import Any, Iterable, Mapping

from qgym.utils import GateEncoder
from qgym.utils.input_validation import check_instance, check_int, check_string


class MachineProperties:
    """``MachineProperties`` is a class to conveniently setup machine properties for the
    ``Scheduling`` environment.
    """

    def __init__(self, n_qubits: int) -> None:
        """
        Init of the MachineProperties class.

        Args:
            n_qubits: Number of qubits of the machine.
        """
        self._n_qubits = check_int(n_qubits, "n_qubits", l_bound=1)
        self._gates: dict[Any, int] = {}
        self._same_start: set[Any] = set()
        self._not_in_same_cycle: dict[Any, set[Any]] = {}

    @classmethod
    def from_mapping(cls, machine_properties: Mapping[str, Any]) -> MachineProperties:
        """Initialize the ``MachineProperties`` class from a ``Mapping`` containing
        valid machines properties.

        Args:
            machine_properties: ``Mapping`` containing valid machine properties.

        Returns:
            Initialized ``MachineProperties`` object with the properties described in
            the `machine_properties` ``Mapping``.
        """
        checked_mp = cls._check_machine_properties_mapping(machine_properties)

        _machine_properties = cls(checked_mp["n_qubits"])
        _machine_properties.add_gates(checked_mp["gates"])
        _machine_properties.add_same_start(checked_mp["same_start"])
        _machine_properties.add_not_in_same_cycle(checked_mp["not_in_same_cycle"])
        return _machine_properties

    @classmethod
    def from_file(cls, filename: str) -> MachineProperties:
        """Load MachineProperties from a JSON file. Not implemented."""
        raise NotImplementedError(
            "Loading machine properties from files is not yet implemented."
        )

    def add_gates(self, gates: Mapping[str, int]) -> MachineProperties:
        """Add gates to the machine properties that should be supported.

        Args:
            gates: ``Mapping`` of gates that the machine can perform as keys, and the
                number of machine cycles (time) as values.

        Returns:
            The ``MachineProperties`` with the added gates.
        """
        check_instance(gates, "gates", Mapping)

        for gate_name, n_cycles in gates.items():
            gate_name = check_string(gate_name, "gate name", lower=True)
            n_cycles = check_int(n_cycles, "n cycles", l_bound=1)
            if gate_name in self._gates:
                msg = f"Gate '{gate_name}' was already given. Overwriting it with the "
                msg += "new value."
                warnings.warn(msg)
            self._gates[gate_name] = n_cycles
            self._not_in_same_cycle[gate_name] = set()
        return self

    def add_same_start(self, gates: Iterable[str]) -> MachineProperties:
        """Add gates that should start in the same cycle, or wait till the previous gate
        is done.

        Args:
            gates: ``Iterable`` of gate names that should start in the same cycle.

        Returns:
            The ``MachineProperties`` with the same start gates.
        """
        check_instance(gates, "gates", Iterable)

        for gate_name in gates:
            gate_name = check_string(gate_name, "gate name", lower=True)
            if gate_name not in self.gates:
                raise ValueError(f"unknown gate '{gate_name}'")
            self._same_start.add(gate_name)
        return self

    def add_not_in_same_cycle(
        self, gates: Iterable[tuple[str, str]]
    ) -> MachineProperties:
        """Add gates that should not start in the same cycle.

        Args:
            gates: ``Iterable`` of tuples of gate names that should not start in the
                same cycle.

        Returns:
            The ``MachineProperties`` with an updated ``not_in_same_cycle`` property.
            The ``not_in_same_cycle`` property is updated according to the input
            `gates`.
        """
        check_instance(gates, "gates", Iterable)
        for gate1, gate2 in gates:
            # Check if the gates are strings and known.
            gate1 = check_string(gate1, "gate", lower=True)
            gate2 = check_string(gate2, "gate", lower=True)
            if gate1 not in self.gates:
                raise ValueError(f"unknown gate '{gate1}'")
            if gate2 not in self.gates:
                raise ValueError(f"unknown gate '{gate2}'")

            self._not_in_same_cycle[gate1].add(gate2)
            self._not_in_same_cycle[gate2].add(gate1)

        return self

    def encode(self) -> GateEncoder:
        """Encode the gates in the machine properties to integer values.

        Returns:
            The GateEncoder used to encode the gates. This GateEncoder can be used to
            decode the gates or encode quantum circuits containing the same gate names
            as in this ``MachineProperties`` object.
        """
        if (
            any(not isinstance(gate, str) for gate in self._gates)
            or any(not isinstance(gate, str) for gate in self._same_start)
            or any(not isinstance(gate, str) for gate in self._not_in_same_cycle)
        ):
            msg = "Gate names of machine properties are not of type str, perhaps they "
            msg = "are already encoded?"
            raise ValueError(msg)

        gate_encoder = GateEncoder().learn_gates(self._gates)

        self._gates = gate_encoder.encode_gates(self._gates)
        self._same_start = gate_encoder.encode_gates(self._same_start)
        self._not_in_same_cycle = gate_encoder.encode_gates(self._not_in_same_cycle)
        return gate_encoder

    @property
    def n_qubits(self) -> int:
        """Return the number of qubits of the machine."""
        return self._n_qubits

    @property
    def gates(self) -> dict[str, int] | dict[int, int]:
        """Return a``Dict`` with the gate names the machine can perform as keys, and the
        number of machine cycles (time) as values.
        """
        return self._gates

    @property
    def n_gates(self) -> int:
        """Return the number of supported gates."""
        return len(self._gates)

    @property
    def same_start(self) -> set[str] | set[int]:
        """Set of gate names that should start in the same cycle, or wait till the
        previous gate is done.
        """
        return self._same_start

    @property
    def not_in_same_cycle(self) -> dict[str, set[str]] | dict[int, set[int]]:
        """Gates that can not start in the same cycle."""
        return self._not_in_same_cycle

    @staticmethod
    def _check_machine_properties_mapping(
        machine_properties: Mapping[str, Any]
    ) -> dict[str, Any]:
        """Check if the given machine properties ``Mapping`` is a valid descriptions of
        the machine properties and returns a ``Dict`` to easily initialize a
        ``MachineProperties`` object.

        Args:
            machine_properties: ``Mapping`` containing the machine properties.

        Raises:
            TypeError: If ``machine_properties`` is not a ``Mapping``.
            ValueError: If there are missing keys in the ``Mapping``.

        Returns:
            ``Dict`` to easily initialize a ``MachineProperties`` object.
        """
        check_instance(machine_properties, "machine_properties", Mapping)
        keys = ["n_qubits", "gates", "machine_restrictions"]
        if not all(key in machine_properties for key in keys):
            msg = "'machine_properties' must have the keys 'n_qubits', 'gates' and "
            msg += "'machine_restrictions'"
            raise ValueError(msg)

        checked_mp = {}
        checked_mp["n_qubits"] = deepcopy(machine_properties["n_qubits"])
        checked_mp["gates"] = deepcopy(machine_properties["gates"])

        machine_restrictions = machine_properties["machine_restrictions"]
        check_instance(machine_restrictions, "machine_restrictions", Mapping)
        keys = ["same_start", "not_in_same_cycle"]
        if not all(key in machine_restrictions for key in keys):
            msg = "'machine_restrictions' must have the keys 'same_start' and "
            msg += "'not_in_same_cycle'"
            raise ValueError(msg)

        checked_mp["same_start"] = deepcopy(machine_restrictions["same_start"])
        checked_mp["not_in_same_cycle"] = []

        for gate1, gate_lst in machine_restrictions["not_in_same_cycle"].items():
            for gate_other in gate_lst:
                checked_mp["not_in_same_cycle"].append((gate1, gate_other))

        return checked_mp

    def __str__(self) -> str:
        """Make a string representation of the machine properties."""
        text = f"{self.__class__.__name__}:\n"
        text += f"n_qubits: {self._n_qubits}\n"
        text += f"gates: {self._gates}\n"
        text += f"same_start: {self._same_start}\n"
        text += f"not_in_same_cycle: {self._not_in_same_cycle}"
        return text

    def __repr__(self) -> str:
        """Make a string representation without endline characters."""
        text = f"{self.__class__.__name__}("
        text += f"n_qubits={self._n_qubits}, "
        text += f"gates={self._gates}, "
        text += f"same_start={self._same_start}, "
        text += f"not_in_same_cycle={self._not_in_same_cycle})"
        return text



"""This module contains a class used for rendering a ``InitialMapping`` environment."""
from __future__ import annotations

from typing import Any

import networkx as nx
import numpy as np
import pygame
from networkx import Graph
from numpy.typing import NDArray

from qgym.envs.initial_mapping.initial_mapping_state import InitialMappingState
from qgym.templates.visualiser import RenderData, Visualiser
from qgym.utils.visualisation.colors import BLACK, BLUE, GRAY, GREEN, RED, WHITE
from qgym.utils.visualisation.typing import Font, Surface
from qgym.utils.visualisation.wrappers import draw_point, draw_wide_line

# pylint: disable=invalid-name


class InitialMappingVisualiser(Visualiser):
    """Visualiser class for the ``InitialMapping`` environment."""

    def __init__(self, render_mode: str, connection_graph: Graph) -> None:
        """Init of the ``InitialMappingVisualiser``.

        Args:
            connection_graph: ``networkx.Graph`` representation of the connection graph.
            render_mode: If 'human' open a ``pygame`` screen visualizing the step. If
                'rgb_array', return an RGB array encoding of the rendered frame on each
                render call.
        """
        # Rendering data
        colors = {
            "nodes": BLUE,
            "basic_edge": BLACK,
            "unused_edge": GRAY,
            "used_edge": GREEN,
            "missing_edge": RED,
            "text": BLACK,
            "background": WHITE,
        }
        self.render_data = RenderData(
            screen=self._start_screen("Mapping Environment", render_mode, (1300, 730)),
            font=self._start_font(),
            colors=colors,
            render_mode=render_mode,
        )

        self.subscreens = self._init_subscreen_rectangles()

        # Save evrything we need to know about the graphs
        self.graphs = {
            "connection": {
                "render_positions": self._get_render_positions(
                    connection_graph, self.subscreens[0]
                ),
                "nodes": connection_graph.nodes,
                "edges": connection_graph.edges,
                "matrix": nx.to_numpy_array(connection_graph),
            },
            "interaction": {"render_positions": {}},
            "mapped": {
                "render_positions": self._get_render_positions(
                    connection_graph, self.subscreens[2]
                ),
            },
        }

    def _init_subscreen_rectangles(
        self, padding: int = 20
    ) -> tuple[pygame.Rect, pygame.Rect, pygame.Rect]:
        """Initialize the ``pygame.Rect`` objects used for drawing the subscreens.

        Args:
            padding: The padding to be used inbetween the subscreens.

        Returns:
            A tuple containing two subscreens.
        """
        small_screen_width = self.screen_width / 2 - 0.5 * padding
        small_screen_height = self.screen_height / 2 - self.header_spacing
        small_screen_shape = (small_screen_width, small_screen_height)

        large_screen_width = self.screen_width / 2 - 0.5 * padding
        large_screen_height = self.screen_height - self.header_spacing
        large_screen_shape = (large_screen_width, large_screen_height)

        screen1_pos = (0, self.header_spacing)
        screen2_pos = (0, self.header_spacing + self.screen_height / 2)
        screen3_pos = (small_screen_width + padding, self.header_spacing)

        subscreen1 = pygame.Rect(screen1_pos, small_screen_shape)
        subscreen2 = pygame.Rect(screen2_pos, small_screen_shape)
        subscreen3 = pygame.Rect(screen3_pos, large_screen_shape)
        return subscreen1, subscreen2, subscreen3

    def render(self, state: InitialMappingState) -> None | NDArray[np.int_]:
        """Render the current state using ``pygame``.

        The upper left screen shows the connection graph. The lower left screen the
        interaction graph. The right screen shows the mapped graph. Gray edges are
        unused, green edges are mapped correctly and red edges need at least on swap.

        Args:
            state: State to render.

        Raises:
            ValueError: When an invalid mode is provided.

        Returns:
            In 'human' mode returns a boolean value encoding whether the ``pygame``
            screen is open. In 'rgb_array' mode returns an RGB array encoding of the
            rendered image.
        """
        self.screen.fill(self.colors["background"])

        mapped_graph = self._get_mapped_graph(
            state.mapping_dict, state.graphs["interaction"]["matrix"]
        )
        self._draw_connection_graph(self.screen)
        self._draw_interaction_graph(
            self.screen, state.steps_done, state.graphs["interaction"]["graph"]
        )
        self._draw_mapped_graph(self.screen, mapped_graph)

        self._draw_header("Connection Graph", self.subscreens[0], self.screen)
        self._draw_header("Interaction Graph", self.subscreens[1], self.screen)
        self._draw_header("Mapped Graph", self.subscreens[2], self.screen)

        return self._display()

    def _get_mapped_graph(
        self, mapping: dict[int, int], interaction_graph_matrix: NDArray[np.float_]
    ) -> nx.Graph:
        """Construct a mapped graph.

        In this graph gray edges are unused, green edges are mapped correctly and red
        edges need at least on swap. This function is used during rendering.

        Args:
            mapping: Mapping dictionary of the state to render.
            interaction_graph_matrix: Interaction graph matrix of the
                current interaction graph.

        Returns:
            Mapped graph.
        """
        # Make the adjacency matrix of the mapped graph
        n_nodes = len(self.graphs["connection"]["nodes"])
        mapped_matrix = np.zeros_like(self.graphs["connection"]["matrix"])
        for map_i, i in mapping.items():
            for map_j, j in mapping.items():
                mapped_matrix[i, j] = interaction_graph_matrix[n_nodes * map_i + map_j]

        # Make a networkx graph of the mapped graph
        graph = nx.Graph()
        for i in range(mapped_matrix.shape[0]):
            graph.add_node(i)

        for i in range(mapped_matrix.shape[0]):
            for j in range(mapped_matrix.shape[1]):
                self._add_colored_edge(graph, mapped_matrix, (i, j))

        # Relabel nodes for drawing
        nodes_mapping = dict(list(enumerate(self.graphs["connection"]["nodes"])))
        graph = nx.relabel_nodes(graph, nodes_mapping)

        return graph

    def _add_colored_edge(
        self,
        graph: nx.Graph,
        mapped_adjacency_matrix: NDArray[np.float_],
        edge: tuple[int, int],
    ) -> None:
        """Give an edge of the graph a color based on the mapping.

        Gray edges are unused, green edges are mapped correctly and red edges need at
        least on swap.

        Args:
            graph: Graph of which the edges must be colored.
            mapped_adjacency_matrix: Adjacency matrix of the mapped graph.
            edge: Edge to color.
        """
        is_connected = self.graphs["connection"]["matrix"][edge] != 0
        is_mapped = mapped_adjacency_matrix[edge] != 0
        if not is_connected and is_mapped:
            graph.add_edge(*edge, color="red")
        if is_connected and is_mapped:
            graph.add_edge(*edge, color="green")
        if is_connected and not is_mapped:
            graph.add_edge(*edge, color="gray")

    def _draw_connection_graph(self, screen: Surface) -> None:
        """Draw the connection graph on subscreen1.

        Args:
            screen: Screen to draw the connection graph on.
        """
        for u, v in self.graphs["connection"]["edges"]:
            pos_u = self.graphs["connection"]["render_positions"][u]
            pos_v = self.graphs["connection"]["render_positions"][v]
            draw_wide_line(screen, self.colors["basic_edge"], pos_u, pos_v)

        for pos in self.graphs["connection"]["render_positions"].values():
            draw_point(screen, pos, self.colors["nodes"])

    def _draw_interaction_graph(
        self, screen: Surface, step: int, interaction_graph: nx.Graph
    ) -> None:
        """Draw the interaction graph on subscreen2.

        Args:
            screen: Screen to draw the interaction graph on.
            step: Current step number.
            interaction_graph: ``networkx.Graph`` representation of the interaction
                graph to draw.
        """
        # If we don't have node positions for the interaction graph for some reason,
        # compute them. If we are at step 0 we should have a new interaction graph, and
        # we should also compute new positions.
        if step == 0 or len(self.graphs["interaction"]["render_positions"]) == 0:
            self.graphs["interaction"]["render_positions"] = self._get_render_positions(
                interaction_graph, self.subscreens[1]
            )

        for u, v in interaction_graph.edges():
            pos_u = self.graphs["interaction"]["render_positions"][u]
            pos_v = self.graphs["interaction"]["render_positions"][v]
            draw_wide_line(screen, self.colors["basic_edge"], pos_u, pos_v)

        for pos in self.graphs["interaction"]["render_positions"].values():
            draw_point(screen, pos, self.colors["nodes"])

    def _draw_mapped_graph(self, screen: Surface, mapped_graph: nx.Graph) -> None:
        """Draw the mapped graph on subscreen3.

        Args:
            screen: Screen to draw the graph on.
            mapped_graph: ``networkx.Graph`` representation of the mapped graph. Each
                edge should have a color attached to it.
        """
        for u, v in mapped_graph.edges():
            pos_u = self.graphs["mapped"]["render_positions"][u]
            pos_v = self.graphs["mapped"]["render_positions"][v]
            if mapped_graph.edges[u, v]["color"] == "red":
                color = self.colors["missing_edge"]
            if mapped_graph.edges[u, v]["color"] == "green":
                color = self.colors["used_edge"]
            if mapped_graph.edges[u, v]["color"] == "gray":
                color = self.colors["unused_edge"]
            draw_wide_line(screen, color, pos_u, pos_v)

        for pos in self.graphs["mapped"]["render_positions"].values():
            draw_point(screen, pos, self.colors["nodes"])

    def _draw_header(self, text: str, subscreen: pygame.Rect, screen: Surface) -> None:
        """Draw a header above a subscreen.

        Args:
            text: Text of the header.
            subscreen: Subscreen to draw the header above.
            screen: Main screen to draw on.
        """
        pygame_text = self.font["header"].render(text, True, self.colors["text"])
        text_center = (subscreen.center[0], subscreen.y - self.header_spacing / 2)
        text_position = pygame_text.get_rect(center=text_center)
        screen.blit(pygame_text, text_position)

    @staticmethod
    def _get_render_positions(
        graph: nx.Graph, subscreen: pygame.Rect
    ) -> dict[Any, NDArray[np.float_]]:
        """Give the positions of the nodes of a graph on a given subscreen.

        Args:
            graph: Graph of which the node positions must be determined.
            ubscreen: the subscreen on which the graph will be drawn.

        Returns:
            A dictionary where the keys are the names of the nodes, and the values are
            the coordinates of these nodes.
        """
        node_positions: dict[Any, NDArray[np.float_]]
        node_positions = nx.spring_layout(graph, threshold=1e-6)

        # Scale and move the node positions to be centered on the subscreen
        for node, position in node_positions.items():
            node_positions[node] = position * 0.45 * subscreen.size + subscreen.center

        return node_positions

    def _start_font(self) -> dict[str, Font]:
        """Start the pygame font.

        Returns:
            pygame fonts for the gate and axis font.
        """
        pygame.font.init()
        return {"header": pygame.font.SysFont("Arial", 30)}

    @property
    def header_spacing(self) -> float:
        """Header spacing."""
        return 30 / 3 * 4



"""This module contains the ``Dict`` space, i.e., a dictionary with fixed strings as
keys and spaces as values.

Usage:
    >>> from qgym.spaces import Box, Dict
    >>> box1 = Box(low=0, high=20, shape=(2, 3), dtype=int)
    >>> box2 = Box(low=-5, high=1, shape=(3, 2), dtype=float)
    >>> Dict(box1=box1, box2=box2)
    Dict(box1:Box([[0 0 0]
     [0 0 0]], [[20 20 20]
     [20 20 20]], (2, 3), int32), box2:Box([[-5. -5.]
     [-5. -5.]
     [-5. -5.]], [[1. 1.]
     [1. 1.]
     [1. 1.]], (3, 2), float64))
"""
from __future__ import annotations

import typing
from typing import Any, Sequence

import gymnasium.spaces
from numpy.random import Generator


class Dict(gymnasium.spaces.Dict):
    """Dictionary of other action/observation spaces for use in RL environments."""

    def __init__(
        self,
        spaces: typing.Dict[str, gymnasium.Space[Any]]
        | Sequence[tuple[str, gymnasium.Space[Any]]]
        | None = None,
        *,
        rng: Generator | None = None,
        **spaces_kwargs: gymnasium.Space[Any],
    ) -> None:
        """Initialize a ``Dict`` space, with string valued keys and spaces inheriting
        from ``gym.Space`` as values.

        Args:
            spaces: Dictionary containing string valued keys and spaces that are to form
            this ``Dict`` space.
            rng: Random number generator to be used in this space, if ``None`` a new one
                will be constructed.
            spaces_kwargs: Spaces that are to form this ``Dict`` space.
        """
        super().__init__(spaces, seed=None, **spaces_kwargs)
        for space in self.spaces.values():
            # override the default behaviour of the gym space
            space._np_random = rng  # pylint: disable=protected-access
